unit FrmDesign;

{

   COPYRIGHT 2014, 2015 Blue Hell / Jan Punter

  Some parts are copyright :

     Author  : Neugls.
     Website : Http://www.neugls.info
     Email   : NeuglsWorkStudio@gmail.com

   these are marked with (*Neugls*)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, Dialogs, ToolWin, StdCtrls, ComCtrls,
  ExtCtrls, Menus, ImgList, Vcl.Buttons, IniFiles, System.UITypes, System.Math,

  JvDesignSurface, JvDesignUtils, JvExControls, JvInspector, JvResources,

  KnobsUtils, Knobs2013, KnobsConversions, SmallForth,

  Globals;

type

  TPixelChanger = function( aColor: TColor): TColor of object;


  TWeatherRec = record
    Direction     : Integer;  // Wind direction 0 .. 359 -> 0 .. 1
    Speed         : Integer;  // Wind speed
    Temperature   : Integer;
    Sunshine      : Integer;  // 0 - 100
    Precipitation : Integer;
    BaroPressure  : Integer;
    Sight         : Integer;
    Humidity      : Integer;  // 0 - 100
  end;


  TWeatherData = class
  private
    FReadFilename     : string;
    FWriteFilename    : string;
    FRawData          : TStringList;
    FData             : array of TWeatherRec;
    FMinSpeed         : Integer;
    FMaxSpeed         : Integer;
    FMinTemperature   : Integer;
    FMaxTemperature   : Integer;
    FMinPrecipitation : Integer;
    FMaxPrecipitation : Integer;
    FMinBaroPressure  : Integer;
    FMaxBaroPressure  : Integer;
    FMinSight         : Integer;
    FMaxSight         : Integer;
    FMinHumidity      : Integer;
    FMaxHumidity      : Integer;
  public
    constructor Create( const aFilename: string);
    destructor  Destroy;                                   override;
    procedure   Calculate;
    procedure   Write( const aFilename: string);
  end;

  TMainForm = class( TForm)
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    ImageList: TImageList;
    MainMenu: TMainMenu;
    MenuItemFile: TMenuItem;
    MenuItemOpen: TMenuItem;
    Save1: TMenuItem;
    MenuItemNew: TMenuItem;
    MenuItemSep1: TMenuItem;
    MenuItemExit: TMenuItem;
    N1: TMenuItem;
    MenuItemDesigenActive: TMenuItem;
    PageControlMain: TPageControl;
    TabSheetModuleDesign: TTabSheet;
    ToolBar: TToolBar;
    but_select: TToolButton;
    but_module: TToolButton;
    but_led: TToolButton;
    but_selector: TToolButton;
    but_knob: TToolButton;
    but_smallknob: TToolButton;
    but_noknob: TToolButton;
    but_display: TToolButton;
    but_fileselector: TToolButton;
    but_indicator: TToolButton;
    but_indicatorbar: TToolButton;
    but_indicatortext: TToolButton;
    but_dataviewer: TToolButton;
    but_input: TToolButton;
    but_output: TToolButton;
    but_textlabel: TToolButton;
    but_editlabel: TToolButton;
    but_box: TToolButton;
    but_image: TToolButton;
    JvDesignScrollBox: TJvDesignScrollBox;
    JvDesignPanel: TJvDesignPanel;
    JvInspector: TJvInspector;
    Splitter: TSplitter;
    TabSheetModuleContents: TTabSheet;
    TimerTest: TTimer;
    TabSheetFonts: TTabSheet;
    FontDialog: TFontDialog;
    PanelFont: TPanel;
    Label5pt: TLabel;
    Label6pt: TLabel;
    Label7pt: TLabel;
    Label8pt: TLabel;
    Label9pt: TLabel;
    ListBoxFonts: TListBox;
    TabSheetForth: TTabSheet;
    PanelDebuggerUI: TPanel;
    TabSheetImages: TTabSheet;
    PanelImages: TPanel;
    OpenDialogBitmap: TOpenDialog;
    ScrollBoxImages: TScrollBox;
    PanelImageControl: TPanel;
    Image: TImage;
    BitBtnSwapRB: TBitBtn;
    BitBtnLoadBitmap: TBitBtn;
    BitBtnSwapRG: TBitBtn;
    BitBtnSwapBG: TBitBtn;
    BitBtnInvertColors: TBitBtn;
    BitBtnMakeTransparent: TBitBtn;
    PanelOutput: TPanel;
    PanelDebuggerControl: TPanel;
    BitBtn1: TBitBtn;
    SplitterDebug: TSplitter;
    PanelOutputControl: TPanel;
    BitBtnClear: TBitBtn;
    PanelOutputClient: TPanel;
    MemoOutput: TMemo;
    EditInput: TEdit;
    PanelDebuggerClient: TPanel;
    MemoDebug: TMemo;
    BitBtnRunForth: TBitBtn;
    BitBtnStep: TBitBtn;
    BitBtnGo: TBitBtn;
    BitBtnStop: TBitBtn;
    BitBtnTestPasCall: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure OpenButtonClick(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure PaletteButtonClick(Sender: TObject);
    procedure MenuItemNewClick(Sender: TObject);
    procedure JvDesignPanelGetAddClass(Sender: TObject; var ioClass: String);
    procedure MenuItemExitClick(Sender: TObject);
    procedure JvDesignPanelSelectionChange(Sender: TObject);
    procedure MenuItemDesigenActiveClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure ListBoxFontsClick(Sender: TObject);
    procedure BitBtnRunForthClick(Sender: TObject);
    procedure EditInputKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure BitBtnClearClick(Sender: TObject);
    procedure BitBtnLoadBitmapClick(Sender: TObject);
    procedure BitBtnSwapRBClick(Sender: TObject);
    procedure BitBtnSwapRGClick(Sender: TObject);
    procedure BitBtnSwapBGClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BitBtnInvertColorsClick(Sender: TObject);
    procedure BitBtnMakeTransparentClick(Sender: TObject);
    procedure BitBtnStepClick(Sender: TObject);
    procedure BitBtnStopClick(Sender: TObject);
    procedure BitBtnGoClick(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
    procedure BitBtnTestPasCallClick(Sender: TObject);
  private
    FDesignClass : string;
    FStickyMode  : Boolean;
    FForth       : TForthThread;
    FBitmap      : TBitmap;
  private
    FIn1  : TSignal; FIn2  : TSignal; FIn3  : TSignal; FIn4  : TSignal;
    FIn5  : TSignal; FIn6  : TSignal; FIn7  : TSignal; FIn8  : TSignal;
    FOut1 : TSignal; FOut2 : TSignal; FOut3 : TSignal; FOut4 : TSignal;
    FOut5 : TSignal; FOut6 : TSignal; FOut7 : TSignal; FOut8 : TSignal;
    FXT   : TFCell;
  private
    procedure   LoadIni;
    procedure   SaveIni;
    procedure   CollectFonts;
    procedure   SelectFont;
  private
    procedure   ClearForthOutput;
    procedure   ClearForthDebugger;
    procedure   ClearForthInput;
    procedure   ForthOutput ( const aMsg: string);
    procedure   DoForthLog  ( const aSender: TObject; const aMsg: string);
    procedure   DoForthDebug( const aSender: TObject; const aMsg: string);
    procedure   DoOnForthTerminated( const aSender: TObject);
    procedure   DoOnForthEndLoading( aSender: PFEngine);
    procedure   DoOnForthInputEmpty( aSender: PFEngine);
    procedure   CallForth;
    procedure   DoOnLoadPFuncs     ( aSender: PFEngine);
    procedure   GetInputs ( aSender: PFEngine; aStack: PFStack);
    procedure   SetOutputs( aSender: PFEngine; aStack: PFStack);
    procedure   RunStopForth;
    procedure   StepForth;
    procedure   StopForth;
    procedure   GoForth;
    procedure   SendForthText( const aValue: string);
    procedure   FocusInput;
  private
    procedure   LoadBitmap;
    function    SwapRB( aColor: TColor): TColor;
    function    SwapRG( aColor: TColor): TColor;
    function    SwapBG( aColor: TColor): TColor;
    function    InvertColors( aColor: TColor): TColor;
    function    MakeBlackTransparent( aColor: TColor): TColor;
    procedure   ChangePixels( aPixelChanger: TPixelChanger);
  private
    function    GetDebugHeight: Integer;
    procedure   SetDebugHeight( aValue: Integer);
  private
    property    DebugHeight : Integer read GetDebugHeight write SetDebugHeight;
  end;

var

  MainForm: TMainForm;



implementation


uses

  JvDesignImp;

// User area


  function EnumFontsProc( var LogFont: TLogFont; var TextMetric: TTextMetric; FontType: Integer; Data: Pointer): Integer; stdcall;
  var
    S    : TStrings;
    Temp : string;
  begin
    S := TStrings( Data);
    Temp := LogFont.lfFaceName;

    if (S.Count = 0) or (AnsiCompareText(S[S.Count-1], Temp) <> 0)
    then S.Add( Temp);

    Result := 1;
  end;

  function CreateFontList: TStringList;
  var
    DC    : HDC;
    LFont : TLogFont;
  begin
    Result := TStringList.Create;
    DC := GetDC( 0);
    FillChar( LFont, sizeof( LFont), 0);
    LFont.lfCharset := DEFAULT_CHARSET;
    EnumFontFamiliesEx( DC, LFont, @ EnumFontsProc, LPARAM( Result), 0);
    ReleaseDC(0, DC);
  end;



{ ========
  TWeatherRec = record
    Direction     : Integer;  // Wind direction 0 .. 359 -> 0 .. 1
    Speed         : Integer;  // Wind speed
    Temperature   : Integer;
    Sunshine      : Integer;  // 0 - 100
    Precipitation : Integer;
    BaroPressure  : Integer;
    Sight         : Integer;
    Humidity      : Integer;  // 0 - 100
  end;

  TWeatherData = class
  private
    FReadFilename     : string;
    FWriteFilename    : string;
    FRawData          : TStringList;
    FData             : array of TWeatherRec;
    FMinSpeed         : Integer;
    FMaxSpeed         : Integer;
    FMinTemperature   : Integer;
    FMaxTemperature   : Integer;
    FMinPrecipitation : Integer;
    FMaxPrecipitation : Integer;
    FMinBarePressure  : Integer;
    FMaxBaroPresssure : Integer;
    FMinSight         : Integer;
    FMaxSight         : Integer;
    FMinHumidity      : Integer;
    FMaxHumidity      : Integer;
  public
}

    constructor TWeatherData.Create( const aFilename: string);
    var
      i : Integer;
    begin
      inherited Create;
      FReadFilename := aFilename;
      FRawData := TStringList.Create;
      if FileExists( FReadFilename)
      then begin
        FRawData.LoadFromFile( FReadFilename);
        i := 0;
        while i < FRawData.Count
        do begin
          if Pos( '#', FRawData[ i]) > 0
          then FRawData.Delete( i)
          else Inc( i);
        end;
      end;
    end;

    destructor  TWeatherData.Destroy; // override;
    begin
      FreeAndNil( FRawData);
      inherited;
    end;

    procedure   TWeatherData.Calculate;
    var
      aLine  : string;
      aParts : TStringList;
      i      : Integer;
      p      : Integer;
    begin
      FMinSpeed         :=   MaxInt;
      FMaxSpeed         := - MaxInt;
      FMinTemperature   :=   MaxInt;
      FMaxTemperature   := - MaxInt;
      FMinPrecipitation :=   MaxInt;
      FMaxPrecipitation := - MaxInt;
      FMinBaroPressure  :=   MaxInt;
      FMaxBaroPressure  := - MaxInt;
      FMinSight         :=   MaxInt;
      FMaxSight         := - MaxInt;
      FMinHumidity      :=   MaxInt;
      FMaxHumidity      := - MaxInt;
      SetLength( FData, FRawData.Count);

      p := 0;
      for aLine in FRawData
      do begin
        aParts := Explode( aLine, ',');
        try
          for i := 0 to aParts.Count - 1
          do aParts[ i] := Trim( aParts[ i]);

          with FData[ p]
          do begin
            Direction := StrToIntDef( aParts[ 2], 0) mod 360;

            Speed := StrToIntDef( aParts[ 3], 0);
            if Speed < FMinSpeed then FMinSpeed := Speed;
            if Speed > FMaxSpeed then FMaxSpeed := Speed;

            Temperature := StrToIntDef( aParts[ 4], 0);
            if Temperature < FMinTemperature then FMinTemperature := Temperature;
            if Temperature > FMaxTemperature then FMaxTemperature := Temperature;

            Sunshine := StrToIntDef( aParts[ 5], 0);

            Precipitation := StrToIntDef( aParts[ 6], 0);
            if Precipitation < 0
            then Precipitation := 0;
            if Precipitation < FMinPrecipitation then FMinPrecipitation := Precipitation;
            if Precipitation > FMaxPrecipitation then FMaxPrecipitation := Precipitation;

            BaroPressure := StrToIntDef( aParts[ 7], 0);
            if BaroPressure < FMinBaroPressure then FMinBaroPressure := BaroPressure;
            if BaroPressure > FMaxBaroPressure then FMaxBaroPressure := BaroPressure;

            Sight := StrToIntDef( aParts[ 8], 0);
            if Sight < FMinSight then FMinSight := Sight;
            if Sight > FMaxSight then FMaxSight := Sight;

            Humidity := StrToIntDef( aParts[ 9], 0);
            if Humidity < FMinHumidity then FMinHumidity := Humidity;
            if Humidity > FMaxHumidity then FMaxHumidity := Humidity;
          end;
        finally
          aParts.Free;
        end;
        Inc( p);
      end;
    end;

    procedure   TWeatherData.Write( const aFilename: string);

      function Scale( const aName: string; aValue, aMin, aMax: Integer): Double;
      begin
        Result := ( aValue - aMin) / ( aMax - aMin);
      end;

    var
      i     : Integer;
      aData : TStringList;
    begin
      FWriteFilename := aFilename;
      aData := TStringList.Create;
      try
        aData.Add( Format( '# MinSpeed         = %d', [ FMinSpeed         ]));
        aData.Add( Format( '# MaxSpeed         = %d', [ FMaxSpeed         ]));
        aData.Add( Format( '# MinTemperature   = %d', [ FMinTemperature   ]));
        aData.Add( Format( '# MaxTemperature   = %d', [ FMaxTemperature   ]));
        aData.Add( Format( '# MinPrecipitation = %d', [ FMinPrecipitation ]));
        aData.Add( Format( '# MaxPrecipitation = %d', [ FMaxPrecipitation ]));
        aData.Add( Format( '# MinBaroPressure  = %d', [ FMinBaroPressure  ]));
        aData.Add( Format( '# MaxBaroPressure  = %d', [ FMaxBaroPressure  ]));
        aData.Add( Format( '# MinSight         = %d', [ FMinSight         ]));
        aData.Add( Format( '# MaxSight         = %d', [ FMaxSight         ]));
        aData.Add( Format( '# MinHumidity      = %d', [ FMinHumidity      ]));
        aData.Add( Format( '# MaxHumidity      = %d', [ FMaxHumidity      ]));
        for i := 0 to Length( FData) - 1
        do begin
          with FData[ i]
          do begin
            aData.Add(
              Format( '%.4f, %.4f, %.4f, %.4f, %.4f, %.4f, %.4f, %.4f',
                [
                  Direction / 360,
                  Scale( 'speed', Speed, FMinSpeed, FMaxSpeed),
                  Scale( 'temperature', Temperature, FMinTemperature, FMaxTemperature),
                  Sunshine / 100,
                  Scale( 'precipitation', Precipitation, FMinPrecipitation, FMaxPrecipitation),
                  Scale( 'baropressure', BaroPressure, FMinBaroPressure, FMaxBAroPressure),
                  Scale( 'sight', Sight, FMinSight, FMaxSight),
                  Scale( 'humidity', Humidity, FMinHumidity, FMaxHumidity)
                ]
              )
            );
          end;
        end;
        aData.SaveToFile( FWriteFilename);
      finally
        aData.Free;
      end;
    end;


  function Color2RGB( aColor: TColor):Integer;
  (*Neugls*)
  begin
    Result := 0;
    Result := Result or ((( aColor shr  0) and $ff) shl 16);
    Result := Result or ((( aColor shr  8) and $ff) shl  8);
    Result := Result or ((( aColor shr 16) and $ff) shl  0);
  end;


type

  TJvInspectorColorItem = class( TJvCustomInspectorItem)
  (*Neugls*)
  protected
    function    GetColorName( const Color: TColor): string;
    procedure   PaintValue( const Color: TColor; const ColorName: string;const ACanvas: TCanvas; const ARect: TRect);
  protected
    procedure   SetFlags( const Value: TInspectorItemFlags);                                                   override;
    procedure   SetRects( const RectKind: TInspectorPaintRect; Value: TRect);                                  override;
    procedure   ButtonClick( Sender: TObject);                                                                 override;
    function    GetDisplayValue: string;                                                                       override;
  public
    constructor Create( const AParent: TJvCustomInspectorItem; const AData: TJvCustomInspectorData);           override;
    procedure   DrawValue( const ACanvas: TCanvas);                                                            override;
  public
    class procedure RegisterAsDefaultItem;
    class procedure UnregisterAsDefaultItem;
  end;

  TJvInspectorControlTypeItem = class( TJvCustomInspectorItem)
  protected
    procedure   SetFlags( const Value: TInspectorItemFlags);                                                   override;
  public
    constructor Create( const AParent: TJvCustomInspectorItem; const AData: TJvCustomInspectorData);           override;
  public
    class procedure RegisterAsDefaultItem;
    class procedure UnregisterAsDefaultItem;
  end;


{ ========
  TJvInspectorColorItem = class( TJvCustomInspectorItem)
  protected
}

    function    TJvInspectorColorItem.GetColorName( const Color: TColor): string;
    (*Neugls*)
    var
      RGBColor:Integer;
    begin
      RGBColor := Color2RGB( Color);
      Result   := '$'+ IntToHex( RGBColor, 6);
    end;

    procedure   TJvInspectorColorItem.PaintValue( const Color: TColor; const ColorName: string;const ACanvas: TCanvas; const ARect: TRect);
    (*Neugls*)
    var
      TH      : Integer;
      BoxRect : TRect;
      bc      : TColor;
      pc      : TColor;
      txtRect : TRect;
    begin
      TH             := Rects[ iprValue].Bottom - Rects[ iprValue].Top - 2;
      BoxRect.Left   := ARect.Left + ( ARect.Bottom - ARect.Top - TH) div 2;
      BoxRect.Top    := ARect.Top  + BoxRect.Left - ARect.Left;
      BoxRect.Right  := BoxRect.Left + TH;
      BoxRect.Bottom := BoxRect.Top  + TH;
      with ACanvas do
      begin
        if Color <> clNone then
        begin
          bc := Brush.Color;
          pc := Pen.Color;
          try
            Brush.Color := Color;
            Pen.Color   := clBlack;
            Rectangle(BoxRect);
          finally
            Pen.Color := pc;
            Brush.Color := bc;
          end;
        end;
       txtRect := ARect;
       txtRect.Left := txtRect.Left + (txtRect.Bottom-txtRect.Top)+ 1;
       TextRect(txtRect, txtRect.Left, BoxRect.Top, ColorName);
      end;
    end;

//  protected

    procedure   TJvInspectorColorItem.SetFlags( const Value: TInspectorItemFlags); // override;
    (*Neugls*)
    begin
      inherited SetFlags( Value + [ iifEditButton, iifEditFixed]);
    end;

    procedure   TJvInspectorColorItem.SetRects( const RectKind: TInspectorPaintRect; Value: TRect); // override;
    (*Neugls*)
    begin
      if RectKind = iprValue
      then Value.Left := Value.Left + (Value.Bottom - Value.Top) + 2;
      inherited SetRects( RectKind, Value);
    end;

    procedure   TJvInspectorColorItem.ButtonClick( Sender: TObject); // override;
    (*Neugls*)
    var
      ColorDlg: TColorDialog;
    begin
      ColorDlg := TColorDialog.Create( nil);
      try
        if ColorDlg.Execute
        then begin
       // DisplayValue := NameForColor( FColor);
          Data.AsOrdinal := ColorDlg.Color;
        end;
      finally
        ColorDlg.Free;
      end;
      inherited;
    end;

    function    TJvInspectorColorItem.GetDisplayValue: string; // override;
    (*Neugls*)
    begin
      Result := GetColorName( Data.AsOrdinal);
    end;

//  public

    constructor TJvInspectorColorItem.Create( const AParent: TJvCustomInspectorItem; const AData: TJvCustomInspectorData); // override;
    (*Neugls*)
    begin
      inherited Create( AParent, AData);
      Flags := [ iifVisible, iifEditButton, iifEditFixed];
    end;

    procedure   TJvInspectorColorItem.DrawValue( const ACanvas: TCanvas); // override;
    (*Neugls*)
    var
      Color     : TColor;
      S         : string;
      ARect     : TRect;
      SafeColor : TColor;
    begin
      Color := clNone;

      if Data = nil
      then S := RsJvInspItemUnInitialized
      else begin
        try
          if not Data.IsInitialized
          then S := RsJvInspItemUnInitialized
          else if not Data.HasValue
          then S := RsJvInspItemNoValue
          else if not Data.IsAssigned
          then S := RsJvInspItemUnassigned
          else begin
            S     := DisplayValue;
            Color := Data.AsOrdinal;
          end;
        except
          on E: Exception
          do S := RsJvInspItemValueException + E.ClassName + ': ' + E.Message;
        end;

        ARect     := Rects[ iprValueArea];
        SafeColor := ACanvas.Brush.Color;

        if Editing
        then ACanvas.Brush.Color := clWindow;

        try
          ACanvas.FillRect( ARect);
          PaintValue( Color, S, ACanvas, ARect);
          if Editing
          then DrawEditor( ACanvas);
        finally
          if Editing
          then ACanvas.Brush.Color := SafeColor;
        end;
      end;
    end;

//  public

    class procedure TJvInspectorColorItem.RegisterAsDefaultItem;
    (*Neugls*)
    begin
      with TJvCustomInspectorData.ItemRegister
      do begin
        if IndexOf( Self) = -1
        then Add( TJvInspectorTypeInfoRegItem.Create( Self, TypeInfo( TColor)));
      end;
    end;

    class procedure TJvInspectorColorItem.UnregisterAsDefaultItem;
    (*Neugls*)
    begin
      TJvCustomInspectorData.ItemRegister.Delete( Self);
    end;


{ ========
  TJvInspectorControlTypeItem = class( TJvCustomInspectorItem)
  protected
}

    procedure   TJvInspectorControlTypeItem.SetFlags( const Value: TInspectorItemFlags); // override;
    begin
      inherited SetFlags( Value + [ iifEditButton, iifEditFixed, iifValueList]);
    end;

//  public

    constructor TJvInspectorControlTypeItem.Create( const AParent: TJvCustomInspectorItem; const AData: TJvCustomInspectorData); // override;
    begin
      inherited Create( AParent, AData);
      Flags := [ iifVisible, iifEditButton, iifEditFixed, iifValueList];
    end;

//  public

    class procedure TJvInspectorControlTypeItem.RegisterAsDefaultItem;
    begin
      with TJvCustomInspectorData.ItemRegister
      do begin
        if IndexOf( Self) = -1
        then Add( TJvInspectorTypeInfoRegItem.Create( Self, TypeInfo( TColor)));
      end;
    end;

    class procedure TJvInspectorControlTypeItem.UnregisterAsDefaultItem;
    begin
      TJvCustomInspectorData.ItemRegister.Delete( Self);
    end;


//  private

const

  sdfs = 'Defaults';
  sdws = 'window';
  hist = 'CmdHistory';

    procedure   TMainForm.LoadIni;
    var
      anIniFile    : TMemIniFile;
      L, T, W, H   : Integer;
      WinState     : TWindowState;
    begin
      anIniFile := TMemIniFile.Create( IniFileName);
      with anIniFile
      do begin
        try
          L                        :=                ReadInteger( sdws, 'Left'          ,          Left           );
          T                        :=                ReadInteger( sdws, 'Top'           ,          Top            );
          W                        :=                ReadInteger( sdws, 'Width'         ,          Width          );
          H                        :=                ReadInteger( sdws, 'Height'        ,          Height         );
          WinState                 := TWindowState(  ReadInteger( sdws, 'WindowState'   , Integer( WindowState )) );
          DebugHeight              :=                ReadInteger( sdws, 'DebugHeight'   ,          DebugHeight    );
          PageControlMain.TabIndex :=                ReadInteger( sdws, 'TabIndex'      , PageControlMain.TabIndex);

          if ( L < 0) or ( L > Screen.Width  - 20) then L := 0;
          if ( T < 0) or ( T > Screen.Height - 20) then T := 0;
          if W > Screen.Width                      then W := Screen.Width  - 20;
          if H > Screen.Height                     then H := Screen.Height - 20;

          SetBounds( L, T, W, H);
          if WinState = wsMaximized
          then WindowState := wsMaximized;
        finally
          Free;
        end;
      end;
    end;

    procedure   TMainForm.SaveIni;
    var
      anIniFile : TMemIniFile;
    begin
      anIniFile := TMemIniFile.Create( IniFileName);
      with anIniFile
      do begin
        try
          EraseSection( sdfs);
          if WindowState = wsNormal
          then begin
            WriteInteger( sdws, 'Left'       , Left       );
            WriteInteger( sdws, 'Top'        , Top        );
            WriteInteger( sdws, 'Width'      , Width      );
            WriteInteger( sdws, 'Height'     , Height     );
          end;
          WriteInteger( sdws, 'WindowState', Integer( WindowState)   );
          WriteInteger( sdws, 'DebugHeight', DebugHeight             );
          WriteInteger( sdws, 'TabIndex'   , PageControlMain.TabIndex);
        finally
          UpdateFile;
          Free;
        end;
      end;
    end;

    procedure   TMainForm.CollectFonts;
    var
      aStringList : TStringList;
    begin
      aStringList := CreateFontList;
      if Assigned( aStringList)
      then begin
        try
          ListBoxFonts.Items := aStringList;
        finally
          aStringList.DisposeOf;
        end;
      end;
    end;

    procedure   TMainForm.SelectFont;
    var
      aFont : TFont;
    begin
      if ListBoxFonts.ItemIndex >= 0
      then begin
        aFont := TFont.Create;
        try
          aFont.Name := ListBoxFonts.Items[ ListBoxFonts.ItemIndex];
          aFont.Size := 5;
          Label5pt.Font := aFont;
          aFont.Size := 6;
          Label6pt.Font := aFont;
          aFont.Size := 7;
          Label7pt.Font := aFont;
          aFont.Size := 8;
          Label8pt.Font := aFont;
          aFont.Size := 9;
          Label9pt.Font := aFont;
        finally
          aFont.DisposeOf;
        end;
      end;
    end;


    procedure   TMainForm.ClearForthOutput;
    begin
      MemoOutput.Clear;
    end;


    procedure   TMainForm.ClearForthDebugger;
    begin
      MemoDebug.Clear;
    end;


    procedure   TMainForm.ClearForthInput;
    begin
      EditInput.Clear;
    end;


    procedure   TMainForm.ForthOutput( const aMsg: string);
    var
      L : string;
      i : Integer;
      p : Integer;
    begin
      L := '';
      p := MemoOutput.Lines.Count - 1;

      if p < 0
      then begin
        MemoOutput.Lines.Add( L);
        p := 0;
      end;

      for i := Low( aMsg) to High( aMsg)
      do begin
        case aMsg[ i] of
          ^L : ClearForthOutput;
          ^M :
            begin
              MemoOutput.Lines[ p] := MemoOutput.Lines[ p] + L;
              L := '';
              MemoOutput.Lines.Add( L);
              Inc( p);
            end;
          else L := L + aMsg[ i];
        end;
      end;

      if L <> ''
      then MemoOutput.Lines[ p] := MemoOutput.Lines[ p] + L;
    end;


    procedure   TMainForm.DoForthLog( const aSender: TObject; const aMsg: string);
    begin
      ForthOutput( aMsg);
    end;


    procedure   TMainForm.DoForthDebug( const aSender: TObject; const aMsg: string);
    begin
      MemoDebug.Lines.Add( aMsg);
    end;


    procedure   TMainForm.DoOnForthTerminated( const aSender: TObject);
    begin
      BitBtnRunForth.Caption := 'Start &4th';
      FXT    := 0;
      FForth := nil;
    end;


    procedure   TMainForm.DoOnForthEndLoading( aSender: PFEngine);
    begin
      DoForthLog( Self, 'Loading ended');
      FXT := FForth.GetForthExecutionToken( 'SynthCall');
    end;


    procedure   TMainForm.DoOnForthInputEmpty( aSender: PFEngine);
    begin
    end;


    procedure   TMainForm.CallForth;
    begin
      // if FXT = 0
      // then FXT := FForth.GetForthExecutionToken( 'SynthCall');
      FForth.ExecuteForthExecutionToken( FXT);
    end;


    procedure   TMainForm.DoOnLoadPFuncs( aSender: PFEngine);
    begin
      aSender.RegisterPFunc( 'GetInputs' , FFL_NONE, GetInputs );
      aSender.RegisterPFunc( 'SetOutputs', FFL_NONE, SetOutputs);
    end;


    procedure   TMainForm.GetInputs( aSender: PFEngine; aStack: PFStack);
    // Build a stackframe with module input values for the forth engine
    begin
      FIn1 := Random;
      FIn2 := 2 * FIn1;
      FIn3 := 3 * FIn1;
      FIn4 := 4 * FIn1;
      FIn5 := 5 * FIn1;
      FIn6 := 6 * FIn1;
      FIn7 := 7 * FIn1;
      FIn8 := 8 * FIn1;
      aStack^.Push( SignalToFractional( FIn1));
      aStack^.Push( SignalToFractional( FIn2));
      aStack^.Push( SignalToFractional( FIn3));
      aStack^.Push( SignalToFractional( FIn4));
      aStack^.Push( SignalToFractional( FIn5));
      aStack^.Push( SignalToFractional( FIn6));
      aStack^.Push( SignalToFractional( FIn7));
      aStack^.Push( SignalToFractional( FIn8));
    end;


    procedure   TMainForm.SetOutputs( aSender: PFEngine; aStack: PFStack);
    // Break down stack frame from forth engine and set 'm as output values for the module
    begin
      FOut8 := FractionalToSignal( aStack^.Pop);
      FOut7 := FractionalToSignal( aStack^.Pop);
      FOut6 := FractionalToSignal( aStack^.Pop);
      FOut5 := FractionalToSignal( aStack^.Pop);
      FOut4 := FractionalToSignal( aStack^.Pop);
      FOut3 := FractionalToSignal( aStack^.Pop);
      FOut2 := FractionalToSignal( aStack^.Pop);
      FOut1 := FractionalToSignal( aStack^.Pop);
      ForthOutput( Format( 'Forth output: %f, %f, %f, %f, %f, %f, %f, %f', [ FOut1, FOut2, FOut3, FOut4, FOut5, FOut6, FOut7, FOut8]));
    end;


    procedure   TMainForm.RunStopForth;
    begin
      if Assigned( FForth)
      then FForth.Abort
      else begin
        FForth :=
          TForthThread.
            Create(
              False              ,          // This is not a Wren evocation
              DoForthLog         ,          // Log                      callback
              DoForthDebug       ,          // Debug                    callback
              DoOnForthTerminated,          // Thread termination       callback
              DoOnForthEndLoading,          // All loading done         callback
              DoOnForthInputEmpty,          // Empty input notification callback
              DoOnLoadPFuncs     ,          // PFunc loader             callback
               1024              ,          // Data    stack size
                256              ,          // Return  stack size
                128              ,          // Control stack size
                128              ,          // Voc     stack size
              65536                         // Dictionary    size
            );
        BitBtnRunForth.Caption := 'Exit &4th';
      end;
    end;


    procedure    TMainForm.StepForth;
    begin
      if Assigned( FForth)
      then FForth.Step( scStepOver);
    end;


    procedure    TMainForm.StopForth;
    begin
      if Assigned( FForth)
      then FForth.Stop;
    end;


    procedure    TMainForm.GoForth;
    begin
      if Assigned( FForth)
      then FForth.Go;
    end;


    procedure   TMainForm.SendForthText( const aValue: string);
    begin
      if Assigned( FForth)
      then FForth.AcceptText( aValue, True);
    end;


    procedure   TMainForm.FocusInput;
    begin
      if EditInput.Visible and EditInput.Enabled and ( PageControlMain.ActivePage = TabSheetForth)
      then EditInput.SetFocus;
    end;


//  private

    procedure   TMainForm.LoadBitmap;
    begin
      with OpenDialogBitmap
      do begin
        if Execute
        then begin
          FBitmap.LoadFromFile( FileName);
          FBitmap.PixelFormat := pf32bit;
          Image.Picture.Bitmap.Assign( FBitmap);
        end;
      end;
    end;


    function    TMainForm.SwapRB( aColor: TColor): TColor;
    begin
      Result := RGB( GetBValue( aColor), GetGValue( aColor), GetRValue( aColor));
    end;


    function    TMainForm.SwapRG( aColor: TColor): TColor;
    begin
      Result := RGB( GetGValue( aColor), GetRValue( aColor), GetBValue( aColor));
    end;


    function    TMainForm.SwapBG( aColor: TColor): TColor;
    begin
      Result := RGB( GetRValue( aColor), GetBValue( aColor), GetGValue( aColor));
    end;


    function    TMainForm.InvertColors( aColor: TColor): TColor;
    begin
      Result := RGB( 255 - GetRValue( aColor), 255 - GetGValue( aColor), 255 - GetBValue( aColor));
    end;


    function    TMainForm.MakeBlackTransparent( aColor: TColor): TColor;
    begin
      Result := TColor((255 - GetRValue( aColor)) shl 24) + aColor;
    end;


    procedure   TMainForm.ChangePixels( aPixelChanger: TPixelChanger);
    var
      i      : Integer;
      j      : Integer;
    begin
      Assert( FBitmap.PixelFormat = pf32bit);

      for i := 0 to FBitmap.Width - 1
      do begin
        for j := 0 to FBitmap.Height - 1
        do FBitmap.Canvas.Pixels[ i, j] := aPixelChanger( FBitmap.Canvas.Pixels[ i, j])
      end;
      Image.Picture.Bitmap.Assign( FBitmap);
    end;


//  private

    function    TMainForm.GetDebugHeight: Integer;
    begin
      Result := PanelDebuggerUI.Height;
    end;


    procedure   TMainForm.SetDebugHeight( aValue: Integer);
    begin
      PanelDebuggerUI.Height := Max( 1, aValue);
      SplitterDebug.Top := PanelDebuggerUI.Top + 1;
    end;



// Delphi area

{$R *.dfm}

procedure TMainForm.EditInputKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Shift = []
  then begin
    if Key = VK_RETURN
    then begin
      SendForthText( EditInput.Text);
      EditInput.Text := '';
    end;
  end;
end;

procedure TMainForm.BitBtnStepClick(Sender: TObject);
begin
  StepForth;
end;

procedure TMainForm.BitBtnStopClick(Sender: TObject);
begin
  StopForth;
end;

procedure TMainForm.BitBtn1Click(Sender: TObject);
begin
  ClearForthDebugger;
  FocusInput;
end;

procedure TMainForm.BitBtnTestPasCallClick(Sender: TObject);
begin
  CallForth;
  FocusInput;
end;

procedure TMainForm.BitBtnClearClick(Sender: TObject);
begin
  ClearForthOutput;
  FocusInput;
end;

procedure TMainForm.BitBtnGoClick(Sender: TObject);
begin
  GoForth;
end;

procedure TMainForm.BitBtnInvertColorsClick(Sender: TObject);
begin
  ChangePixels( InvertColors);
end;

procedure TMainForm.BitBtnLoadBitmapClick(Sender: TObject);
begin
  LoadBitmap;
end;

procedure TMainForm.BitBtnRunForthClick(Sender: TObject);
begin
  RunStopForth;
  FocusInput;
end;

procedure TMainForm.BitBtnMakeTransparentClick(Sender: TObject);
begin
  ChangePixels( MakeBlackTransparent);
end;

procedure TMainForm.BitBtnSwapBGClick(Sender: TObject);
begin
  ChangePixels( SwapBG);
end;

procedure TMainForm.BitBtnSwapRBClick(Sender: TObject);
begin
  ChangePixels( SwapRB);
end;

procedure TMainForm.BitBtnSwapRGClick(Sender: TObject);
begin
  ChangePixels( SwapRG);
end;

procedure TMainForm.ListBoxFontsClick(Sender: TObject);
begin
  SelectFont;
end;

procedure TMainForm.MenuItemExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.MenuItemDesigenActiveClick(Sender: TObject);
begin
  JvDesignPanel.Active := not JvDesignPanel.Active;
  MenuItemDesigenActive.Checked := JvDesignPanel.Active;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if Assigned( FForth)
  then FForth.Abort;
  SaveIni;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  OpenDialog.InitialDir := ExtractFilePath(Application.ExeName);
  SaveDialog.InitialDir := OpenDialog.InitialDir;
  JvDesignPanel.Surface.MessengerClass := TJvDesignWinControlHookMessenger;
// JvDesignPanel.Surface.MessengerClass := TJvDesignDesignerMessenger;
  JvDesignPanel.Surface.Active  := True;
  MenuItemDesigenActive.Checked := JvDesignPanel.Surface.Active;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat     := pf32bit;
  FBitmap.TransparentMode := tmFixed;
  FBitmap.Transparent     := False;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil( FBitmap);
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  ClearForthOutput;
  ClearForthDebugger;
  ClearForthInput;
  LoadIni;
  CollectFonts;
  RunStopForth;
  FocusInput;
end;

procedure TMainForm.JvDesignPanelGetAddClass(Sender: TObject;
  var ioClass: String);
begin
  ioClass := FDesignClass;

  if not FStickyMode
  then begin
    FDesignClass := '';
    but_select.Down := True;
  end;
end;

procedure TMainForm.PaletteButtonClick(Sender: TObject);
const
  cClasses: array[ 0 .. 18] of string =
  (
    ''                   ,
    'TKnobsModule'       ,
    'TKnobsLed'          ,
    'TKnobsSelector'     ,
    'TKnobsKnob'         ,
    'TKnobsSmallKnob'    ,
    'TKnobsNoKnob'       ,
    'TKnobsDisplay'      ,
    'TKnobsFileSelector' ,
    'TKnobsIndicator'    ,
    'TKnobsIndicatorBar' ,
    'TKnobsIndicatorText',
    'TKnobsDataViewer'   ,
    'TKnobsInput'        ,
    'TKnobsOutput'       ,
    'TKnobsTextLabel'    ,
    'TKnobsEditLabel'    ,
    'TKnobsBox'          ,
    'TImage'
  );
begin
  FStickyMode  := GetKeyState( VK_SHIFT) < 0;
  FDesignClass := cClasses[TControl( Sender).Tag];
end;

procedure TMainForm.MenuItemNewClick(Sender: TObject);
begin
  JvDesignPanel.Clear;
end;

procedure TMainForm.OpenButtonClick(Sender: TObject);
begin
  if OpenDialog.Execute
  then JvDesignPanel.LoadFromFile( OpenDialog.Filename);
end;

procedure TMainForm.SaveButtonClick(Sender: TObject);
begin
  if SaveDialog.Execute
  then JvDesignPanel.SaveToFile( SaveDialog.Filename);
end;

procedure TMainForm.JvDesignPanelSelectionChange(Sender: TObject);
begin
  if Length( JvDesignPanel.Surface.Selected) > 0
  then begin
    // JvInspector.AddComponent( JvDesignPanel.Surface.Selected[ 0]);
    JvInspector.InspectObject := JvDesignPanel.Surface.Selected[ 0];
  end
  else JvInspector.InspectObject := nil;
end;


initialization


  RegisterClass( TKnobsModule       );
  RegisterClass( TKnobsLed          );
  RegisterClass( TKnobsSelector     );
  RegisterClass( TKnobsKnob         );
  RegisterClass( TKnobsSmallKnob    );
  RegisterClass( TKnobsNoKnob       );
  RegisterClass( TKnobsDisplay      );
  RegisterClass( TKnobsFileSelector );
  RegisterClass( TKnobsIndicator    );
  RegisterClass( TKnobsIndicatorBar );
  RegisterClass( TKnobsIndicatorText);
  RegisterClass( TKnobsDataViewer   );
  RegisterClass( TKnobsInput        );
  RegisterClass( TKnobsOutput       );
  RegisterClass( TKnobsTextLabel    );
  RegisterClass( TKnobsEditLabel    );
  RegisterClass( TKnobsBox          );
  RegisterClass( TImage             );

  TJvInspectorColorItem      .RegisterAsDefaultItem;
  TJvInspectorControlTypeItem.RegisterAsDefaultItem;

end.
